


"""This module consists of functions which create the
admin css styles
"""

import copy, colorsys


def get_colours(R, G, B):

    colours = {}

    w3_theme_background_color = int_hex(R, G, B)

    colours["w3_theme_background_color" ] = w3_theme_background_color
    colours["w3_theme_color" ] = textcolor(R, G, B)

    w3_text_theme_color, r, g, b = other_hue_change(R, G, B)
    colours["w3_text_theme_color"]  = w3_text_theme_color
    colours["w3_border_theme_color"] = colours["w3_text_theme_color"]

    w3_theme_light_background_color, r, g, b = light_color(R, G, B)
    colours["w3_theme_light_background_color" ] = w3_theme_light_background_color
    colours["w3_theme_light_color"] = textcolor(r, g, b)

    w3_hover_theme_hover_background_color, r, g, b = hue_change(R, G, B)
    colours["w3_hover_theme_hover_background_color"] = w3_hover_theme_hover_background_color
    colours["w3_hover_theme_hover_color"] = textcolor(r, g, b)
    colours["w3_hover_text_theme_color"], r, g, b = hue_change(r, g, b)
    colours["w3_hover_border_theme_hover_color"] = colours["w3_hover_text_theme_color"]
 

    w3_theme_dark_background_color, r, g, b = dark_color(R, G, B)
    colours["w3_theme_dark_background_color" ] = w3_theme_dark_background_color
    colours["w3_theme_dark_color"] = textcolor(r,g,b)


    w3_theme_action_background_color, r, g, b = dark_color(r, g, b)
    colours["w3_theme_action_background_color" ] = w3_theme_action_background_color
    colours["w3_theme_action_color"] = textcolor(r,g,b)

    w3_theme_l1_background_color, r, g, b = lighter_color(R, G, B)
    colours["w3_theme_l1_background_color" ] = w3_theme_l1_background_color
    colours["w3_theme_l1_color"] = textcolor(r, g, b)

    w3_theme_l2_background_color, r, g, b = lighter_color(r, g, b)
    colours["w3_theme_l2_background_color" ] = w3_theme_l2_background_color
    colours["w3_theme_l2_color"] = textcolor(r, g, b)

    w3_theme_l3_background_color, r, g, b = lighter_color(r, g, b)
    colours["w3_theme_l3_background_color" ] = w3_theme_l3_background_color
    colours["w3_theme_l3_color"] = textcolor(r, g, b)

    w3_theme_l4_background_color, r, g, b = lighter_color(r, g, b)
    colours["w3_theme_l4_background_color" ] = w3_theme_l4_background_color
    colours["w3_theme_l4_color"] = textcolor(r, g, b)

    w3_theme_l5_background_color, r, g, b = lighter_color(r, g, b)
    colours["w3_theme_l5_background_color" ] = w3_theme_l5_background_color
    colours["w3_theme_l5_color"] = textcolor(r, g, b)

    w3_theme_d1_background_color, r, g, b = darker_color(R, G, B)
    colours["w3_theme_d1_background_color" ] = w3_theme_d1_background_color
    colours["w3_theme_d1_color"] = textcolor(r, g, b)

    w3_theme_d2_background_color, r, g, b = darker_color(r, g, b)
    colours["w3_theme_d2_background_color" ] = w3_theme_d2_background_color
    colours["w3_theme_d2_color"] = textcolor(r, g, b)

    w3_theme_d3_background_color, r, g, b = darker_color(r, g, b)
    colours["w3_theme_d3_background_color" ] = w3_theme_d3_background_color
    colours["w3_theme_d3_color"] = textcolor(r, g, b)

    w3_theme_d4_background_color, r, g, b = darker_color(r, g, b)
    colours["w3_theme_d4_background_color" ] = w3_theme_d4_background_color
    colours["w3_theme_d4_color"] = textcolor(r, g, b)

    w3_theme_d5_background_color, r, g, b = darker_color(r, g, b)
    colours["w3_theme_d5_background_color" ] = w3_theme_d5_background_color
    colours["w3_theme_d5_color"] = textcolor(r, g, b)

    return colours




def add_styles(style1, style2):
    """merges a copy of style2 into style1 - style1 changed in-place, returns None"""
    c_style2 = copy.deepcopy(style2)
    for key in c_style2:
        style1[key] = c_style2[key]

def textcolor(r, g, b):
    "Given a background color, suggests a textcolor and returns text-hexstring"
    if r>250 and g>250 and b>250:
        # If background is white, use black text
        return '#000000'
    elif r<100 and g<100 and b>150:
        # if background is blue, use white text
        return '#ffffff'
    elif r<150 and g<150 and b>250:
        # if background is blue, use white text
        return '#ffffff'

    h,s,l = rgb_hsl(r, g, b)
    # get a contrasting tone for text
    if l>0.8:
        # for light background, make text dark, but same hue
        opl = 0.2
    elif l>0.5:
        # for lightish background, make text black
        return '#000000'
    elif l>0.2:
        # for darkish background, make text white
        return '#ffffff'
    else:
        # for dark background, have lighter text
        opl = 0.8
    # Choose same saturation
    new_r, new_g, new_b = hsl_rgb(h, s, opl)
    return int_hex(new_r, new_g, new_b) 

def light_color(r, g, b):
    "Given a background color, suggests a lighter background color and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_l = (l + 1.0)/2.0
    r, g, b = hsl_rgb(h, s, new_l)
    return int_hex(r,g,b), r, g, b

def lighter_color(r, g, b):
    "Given a background color, suggests a lighter background color and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_l = (l + 2.0)/3.0
    r, g, b = hsl_rgb(h, s, new_l)
    return int_hex(r,g,b), r, g, b

def dark_color(r, g, b):
    "Given a background color, suggests a darker background color and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_l = l/2.0
    r, g, b = hsl_rgb(h, s, new_l)
    return int_hex(r,g,b), r, g, b

def darker_color(r, g, b):
    "Given a background color, suggests a darker background color and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_l = 2.0*l/3.0
    r, g, b = hsl_rgb(h, s, new_l)
    return int_hex(r,g,b), r, g, b

def hue_change(r, g, b):
    "Given a background color, suggests a hue change color, and l of 0.5 and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_h = h+0.2
    if new_h>1:
        new_h = new_h - 1
    r, g, b = hsl_rgb(new_h, s, 0.5)
    return int_hex(r,g,b), r, g, b

def other_hue_change(r, g, b):
    "Given a background color, suggests a hue change color, and l of 0.5 and returns (hexstring, r,g,b)"
    h,s,l = rgb_hsl(r, g, b)
    new_h = h-0.2
    if new_h<0:
        new_h = new_h + 1
    r, g, b = hsl_rgb(new_h, s, 0.5)
    return int_hex(r,g,b), r, g, b


def int_hex(r, g, b):
    "converts rgb to a hex string"
    hex_value = '#'
    hexr = hex(r)[2:]
    if len(hexr) == 1:
        hex_value += "0"+hexr
    else:
        hex_value += hexr
    hexg = hex(g)[2:]
    if len(hexg) == 1:
        hex_value += "0"+hexg
    else:
        hex_value += hexg
    hexb = hex(b)[2:]
    if len(hexb) == 1:
        hex_value += "0"+hexb
    else:
        hex_value += hexb
    return hex_value

def hex_int(hexstring):
    "converts #hexdigits to r,g,b integer colours"
    if not hexstring:
        return 0,0,0
    hexstring = hexstring.strip('#')
    r = int(hexstring[0:2], base=16)
    g = int(hexstring[2:4], base=16)
    b = int(hexstring[4:6], base=16)
    return r, g, b

def rgb_hsl(r, g, b):  # rgb 0 to 255
    "converts rgb values to hsl"
    h,l,s = colorsys.rgb_to_hls(r/255.0, g/255.0, b/255.0)
    return h,s,l

def hsl_rgb(h, s, l):  # hsl from 0 to 1
    "converts hsl values to rgb"
    r,g,b = colorsys.hls_to_rgb(h, l, s)
    r = round(255 * r)
    g = round(255 * g)
    b = round(255 * b)
    if r>255: r=255
    if g>255: g=255
    if b>255: b=255
    return r,g,b







"""Functions that read a json string to create sections, pages and folders"""

import os, json, collections

from ..ski import skiboot, read_json, dump_project
from ..ski.excepts import ServerError

from . import project_loaded, item_info, ident_exists, get_proj_section, get_proj_page


def project_json_file(project):
    "Given a project name, returns the file path of the project.json file"
    return skiboot.project_json(project)


def item_outline(project, pagenumber, section_name, location):
    """Returns a list defining the item and its contents. Such as ['Part', dictionary] for Part,
      ['ClosedPart', dictionary] for ClosedPart etc.,

       The project must be currently loaded as either the root project or a sub-project, and the part must exist in the page.
       pagenumber and section_name are mutually exclusive, one must be None"""
    # item is either in a page or a section
    if (pagenumber is None) and (section_name is None):
        raise ServerError("Page and section both missing")
    if section_name:
        if pagenumber is not None:
            raise ServerError("Item cannot be in both a page and a section")
        # item is in a section
        proj, section = get_proj_section(project, section_name)
        item = section.location_item(location)
    else:
        # item is in a page
        proj, page = get_proj_page(project, pagenumber)
        item = page.location_item(location)
    if item is None:
        raise ServerError("Item not recognised")
    if hasattr(item, 'outline'):
        return item.outline(project)
    else:
        # must be a text string
        return ['Text', str(item)]


def create_section(project, section_name, json_data):
    """Builds the section from the given json string or ordered dictionary, and adds it to project


       Given a project name, a section name and a json string (or ordered dictionary) describing the section,
       this function creates a new section (or overwrites a section with the same name).
       The project must be currently loaded as either the root project or a sub-project.
"""
    # raise error if invalid project
    project_loaded(project)
    try:
        read_json.create_section(project, section_name, json_data)
    except Exception:
        raise ServerError("Unable to create section")


def section_outline(project, section_name):
    """Given a project name, and a section name, returns an outline list defining the section.

       The project must be currently loaded as either the root project or a sub-project,
       and the section must exist in the project."""
    # raise error if invalid project
    project_loaded(project)
    proj = skiboot.getproject(project)
    section = proj.section(section_name, makecopy=False)
    if section is None:
        raise ServerError(message="Section not found")
    return section.outline(project)


def container_outline(project, pagenumber, section_name, widget_name, container):
    """Builds an outline list from the widget container, ServerError if not found"""
    # widget is either in a page or a section
    if (pagenumber is None) and (section_name is None):
        raise ServerError("Page and section both missing")
    if section_name:
        if pagenumber is not None:
            raise ServerError("Widget cannot be in both a page and a section")
        # widget is in a section
        proj, section = get_proj_section(project, section_name)
        widget = section.widgets[widget_name]
    else:
        # widget is in a page
        proj, page = get_proj_page(project, pagenumber)
        widget = page.widgets[widget_name]
    if widget is None:
        raise ServerError("widget not recognised")
    parttext, partdict = widget.outline(project)
    container_name = "container_%s" % container
    # partdict[container_name] is a list of parts in the container, for compatability
    # with other outlines, set this into a dictionary with key 'parts'
    container_dict = collections.OrderedDict()
    container_dict["parts"] = partdict[container_name]
    return ["Container", container_dict]



def create_page(project, parentnumber, pagenumber, page_name, page_brief, json_data):
    """Builds a page from the given json string / ordered dictionary, and adds it to project

       Given a project name, a parent folder number, a new page number, a page name, a brief description
       and a json string (or ordered dictionary) describing the page, it creates a new page with the new page number.
       The number should not already exist in the project."""
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(pagenumber, int):
        raise ServerError(message="pagenumber is not an integer")
    if not isinstance(parentnumber, int):
        raise ServerError(message="parentnumber is not an integer")
    if ident_exists(project, pagenumber):
        raise ServerError(message = "An item with this ident number already exists")
    parentinfo = item_info(project, parentnumber)
    if not parentinfo:
        raise ServerError(message = "The parent folder with this ident number has not been found")
    if parentinfo.item_type != "Folder":
        raise ServerError(message = "The parent with this ident number is not a folder")
    if isinstance(json_data, str):
        page_dict = json.loads(json_data, object_pairs_hook=collections.OrderedDict)
    else:
        page_dict = json_data
    # create the page
    if "SVG" in page_dict:
        read_json.create_svgpage(project, parentnumber, pagenumber, page_name, page_brief, page_dict)
        return
    elif "TemplatePage" in page_dict:
        read_json.create_templatepage(project, parentnumber, pagenumber, page_name, page_brief, page_dict)
        return
    raise ServerError(message = "Invalid JSON file")


def page_to_OD(project, pagenumber):
    """Given a project name, and a page number, returns an Ordered Dictionary defining the page.

       The project must be currently loaded as either the root project or a sub-project,
       and the page must exist in the project."""
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(pagenumber, int):
        raise ServerError(message="pagenumber is not an integer")
    page = skiboot.from_ident(pagenumber, project)
    if page is None:
        raise ServerError(message="Page not found")
    if page.page_type == 'SVG':
        page_dict = dump_project.svg_to_OD(project, page)
    elif page.page_type == 'TemplatePage':
        page_dict = dump_project.templatepage_to_OD(project, page)
    else:
        raise ServerError(message="The given pagenumber is not that of an svg or template page")
    return page_dict


def page_to_json(project, pagenumber, indent=0):
    """Given a project name, and a page number, returns a json string defining the page.

       The project must be currently loaded as either the root project or a sub-project,
       and the page must exist in the project.
       The indent parameter should be the number of indentation spaces to use when formatting the string."""
    page_dict = page_to_OD(project, pagenumber)
    return json.dumps(page_dict, indent=indent, separators=(',', ':'))


def create_folder(project, parentnumber, addition_number, folder_name, restricted, json_data):
    """Builds a folder and contents from the given json string / ordered dictionary, and adds it to project
       returns top folder_number


       Given a project name, a parent folder number, an addition_number, a folder name, a True or False restricted value
       and a json string (or ordered dictionary) describing the folder and its contents, it creates a new folder and contents.
       The folder name should not already exist in the parent folder.
       The json string is typically derived from a previously saved JSON file, and each folder and page within it will already
       have an ident number. The 'addition_number' argument of this function is added to each ident number, and should take the
       new ident numbers outside the range of existing numbers within the project.
"""
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(addition_number, int):
        raise ServerError(message="addition number is not an integer")
    if not isinstance(parentnumber, int):
        raise ServerError(message="parentnumber is not an integer")
    parentinfo = item_info(project, parentnumber)
    if not parentinfo:
        raise ServerError(message = "The parent folder with this ident number has not been found")
    if parentinfo.item_type != "Folder":
        raise ServerError(message = "The parent with this ident number is not a folder")
    if parentinfo.restricted:
        restricted = True
    # create the folder
    ident = read_json.create_folder(project, parentnumber, addition_number, folder_name, restricted, json_data)
    return ident.num


def folder_to_OD(project, foldernumber):
    """Given a project name, and a folder number, returns an Ordered Dictionary defining the folder and its contents.

       The project must be currently loaded as either the root project or a sub-project,
       and the folder must exist in the project."""
    # raise error if invalid project
    project_loaded(project)
    if not isinstance(foldernumber, int):
        raise ServerError(message="foldernumber is not an integer")
    folder = skiboot.from_ident(foldernumber, project)
    if folder is None:
        raise ServerError(message="Folder not found")
    if folder.page_type != 'Folder':
        raise ServerError(message="The given foldernumber is not that of a Folder")
    return dump_project.folder_to_OD(project, folder)


def folder_to_json(project, foldernumber, indent=0):
    """Given a project name, and a folder number, returns a json string defining the folder and its contents.

       The project must be currently loaded as either the root project or a sub-project,
       and the folder must exist in the project.
       The indent parameter should be the number of indentation spaces to use when formatting the string."""
    folder_dict = folder_to_OD(project, foldernumber)
    return json.dumps(folder_dict, indent=indent, separators=(',', ':'))


def project_to_OD(project):
    """For the given project, returns an Ordered Dictionary, which could be large.

       The dictionary describes the project pages and widgets."""
    # raise error if invalid project
    project_loaded(project)
    return dump_project.project_to_OD(project)


def project_to_json(project, save_to_file=True, indent=0):
    """For the given project, creates a json file, or returns a json string, which could be large.

       If save_to_file is True, this function creates the file project.json under the projectfiles/project/data directory, and returns None.
       If save_to_file is False, the function returns a json string. This string describes the project pages and widgets.
       The indent parameter should be the number of indentation spaces to use when formatting the string."""

    project_dict = project_to_OD(project)

    if save_to_file:
        # write out the project dictionary to a json file
        filepath = project_json_file(project)
        with open(filepath, 'w') as fp:
            json.dump(project_dict, fp, indent=indent)
        return

    return json.dumps(project_dict, indent=indent, separators=(',', ':'))


def get_defaults_from_file(projectfiles, project, key=None):
    """Obtains values from the defaults.json file of the given project.
       This can be called before a WSGIApplication object is created
       since it uses the projectfiles information

        If a key is given, it returns the value of that key,
        otherwise it returns the entire dictionary of the JSON file.
        If no value is found, returns None."""
    defaults = {}
    defaultsfile = os.path.join(projectfiles, project, 'data', 'defaults.json')
    if os.path.isfile(defaultsfile):
        # load defaultsfile into defaults
        with open(defaultsfile, 'r') as fp:
            defaults = json.load(fp, object_pairs_hook=collections.OrderedDict)
    else:
        return
    if not defaults:
        return
    if not key:
        return defaults
    if key in defaults:
        return defaults[key]



def get_defaults(project, key=None):
    """Obtains values from the defaults.json file of the given project.

        If a key is given, it returns the value of that key,
        otherwise it returns the entire dictionary of the JSON file.
        If no value is found, returns None."""
    defaults = {}
    defaultsfile = skiboot.project_defaults(proj_ident=project)
    if os.path.isfile(defaultsfile):
        # load defaultsfile into defaults
        with open(defaultsfile, 'r') as fp:
            defaults = json.load(fp, object_pairs_hook=collections.OrderedDict)
    else:
        return
    if not defaults:
        return
    if not key:
        return defaults
    if key in defaults:
        return defaults[key]


def get_widget_default_field_value(project, widg_module, widg_class, widg_field):
    """Returns the default value to set into a widget field argument, returns '' if not found

       When a widget is created or edited, certain fields can be set with defaults - currently CSS classes and CSS Styles.
       You have the option to save these defaults for the widget field.
       These defaults are held in the project 'defaults.json' file.
       This function reads the defaults.json file and returns the default string value for the given widget and field.
       To specify the widget and field, you need to supply the project name, and:

       widg_module is the module name of the widget.
       widg_class is the class name of the widget
       widg_field is the field name of the widget
       If the default value is not found, the function returns an empty string."""
    if not (widg_module and widg_class and widg_field):
        return ''
    widg_defaults_dict = get_defaults(project, key='widgets')
    if widg_defaults_dict is None:
        return ''
    # widg_defaults_dict is a dictionary of keys being module names, values being another dictionary
    if widg_module in widg_defaults_dict:
        widg_module_dict = widg_defaults_dict[widg_module]
    else:
        return ''
    # widg_module_dict is a dictionary of keys being widget class names in the module, values being another dictionary
    if widg_class in widg_module_dict:
        widg_class_dict = widg_module_dict[widg_class]
    else:
        return ''
    # widg_class_dict is a dictionary of keys being widget field arg names in the module, values being the default field values
    if widg_field in widg_class_dict:
        default_field_value = widg_class_dict[widg_field]
    else:
        return ''
    if default_field_value:
        return default_field_value
    return ''


def set_defaults(project, key, value):
    "Sets a key:value pair into the defaults.json file for the project."
    defaults = get_defaults(project)
    if not defaults:
        defaults = {}
    defaults[key] = value
    save_defaults(project, defaults)


def save_defaults(project, defaults):
    "Given a defaults dictionary, saves the entire dictionary as the defaults.json file for the project."
    # create defaults.json
    defaults_json_filename = skiboot.project_defaults(project)
    # write out the project defaults dictionary to a json file
    ordered_defaults = collections.OrderedDict(sorted(defaults.items(), key=lambda t: t[0]))
    # put the 'widgets' key dictionary in order
    modules = ordered_defaults['widgets']
    ordered_modules = collections.OrderedDict(sorted(modules.items(), key=lambda t: t[0]))
    ordered_defaults['widgets'] = ordered_modules
    # modules are in correct order, for each module, sort its classes
    for module, classdict in ordered_modules.items():
        ordered_modules[module] = collections.OrderedDict(sorted(classdict.items(), key=lambda t: t[0]))
    # modules and classes are in correct order, for each class, sort its fields
    for classdict in ordered_modules.values():
        for widgclass, widgfield in classdict.items():
            classdict[widgclass] = collections.OrderedDict(sorted(widgfield.items(), key=lambda t: t[0]))
    with open(defaults_json_filename, 'w') as fp:
        json.dump(ordered_defaults, fp, indent=0)


def save_widget_default_field_value(project, widg_module, widg_class, widg_field, value):
    """Saves a default value string for a given widget field (currently CSS classes and CSS styles) in the project 'defaults.json' file.

       To specify the widget and field, you need to supply the project name, and:

       widg_module is the module name of the widget.
       widg_class is the class name of the widget
       widg_field is the field name of the widget

       The function returns True if ok, False if not."""
    if not (widg_module and widg_class and widg_field):
        return False
    widg_defaults_dict = get_defaults(project, key='widgets')
    if widg_defaults_dict is None:
        widg_defaults_dict = {}
    # widg_defaults_dict is a dictionary of keys being module names, values being another dictionary
    if widg_module in widg_defaults_dict:
        widg_module_dict = widg_defaults_dict[widg_module]
    else:
        widg_module_dict = {}
        widg_defaults_dict[widg_module] = widg_module_dict
    # widg_module_dict is a dictionary of keys being widget class names in the module, values being another dictionary
    if widg_class in widg_module_dict:
        widg_class_dict = widg_module_dict[widg_class]
    else:
        widg_class_dict = {}
        widg_module_dict[widg_class] = widg_class_dict
    # widg_class_dict is a dictionary of keys being widget field arg names in the module, values being the default field value
    widg_class_dict[widg_field] = value
    try:
        set_defaults(project, key='widgets', value=widg_defaults_dict)
    except Exception:
        return False
    return True








"""
named tuples used by other modules
"""

from collections import namedtuple

VersionInfo = namedtuple('VersionInfo', ['skipole', 'project'])

ProjectInfo = namedtuple('ProjectInfo', ['project', 'version', 'brief', 'path', 'default_language', 'subprojects', 'json_path', 'main_path', 'static_path', 'data_path'])

ItemInfo = namedtuple('ItemInfo', ['project', 'project_version', 'itemnumber', 'item_type', 'name', 'brief', 'path', 'label_list', 'change', 'parentfolder_number', 'restricted'])

PartInfo = namedtuple('PartInfo', ['project', 'pagenumber', 'page_part', 'section_name', 'name', 'location', 'part_type', 'brief'])

PageInfo = namedtuple('PageInfo', ['name', 'number', 'restricted', 'brief', 'item_type', 'responder', 'enable_cache', 'change', 'parentfolder_number'])

FolderInfo = namedtuple('FolderInfo', ['name', 'number', 'restricted', 'brief', 'contains_pages', 'contains_folders', 'change'])

WidgetInfo = namedtuple('WidgetInfo', ['project', 'pagenumber', 'section_name', 'name', 'location', 'containers', 'display_errors', 'brief'])

PageElement = namedtuple('PageElement', ['project', 'pagenumber', 'pchange', 'location', 'page_part', 'part_type', 'tag_name', 'brief', 'hide_if_empty', 'attribs'])

PageTextBlock = namedtuple('PageTextBlock', ['project', 'pagenumber', 'pchange', 'location', 'textref', 'tblock_project', 'failmessage', 'escape', 'linebreaks'])

WidgFieldInfo = namedtuple('WidgFieldInfo', ['tuple_widgfield', 'str_widgfield', 'str_comma_widgfield'])

PlaceHolderInfo = namedtuple('PlaceHolderInfo', ['project', 'pagenumber', 'section_name', 'alias', 'brief', 'multiplier', 'mtag'])

SectionElement = namedtuple('SectionElement', ['project', 'section_name', 'schange', 'location', 'part_type', 'tag_name', 'brief', 'show', 'hide_if_empty', 'attribs'])

SectionTextBlock = namedtuple('SectionTextBlock', ['project', 'section_name', 'schange', 'location', 'textref', 'tblock_project', 'failmessage', 'escape', 'linebreaks'])


WidgetDescription = namedtuple('WidgetDescription', ['modulename', 'classname', 'brief', 'fields', 'containers', 'illustration',
                                                     'fields_single', 'fields_list', 'fields_table', 'fields_dictionary', 'parent_widget', 'parent_container'])
# 'fields' is a list of field args
# 'containers' is the number of containers in the widget, 0 for none


FieldDescription = namedtuple('FieldDescription', ['field_arg', 'field_type', 'valdt', 'jsonset', 'cssclass', 'cssstyle'])

ContainerInfo = namedtuple('ContainerInfo', ['container', 'empty'])


ResponderInfo = namedtuple('ResponderInfo', ['responder',
                                             'module_name',
                                             'widgfield_required',
                                             'widgfield',
                                             'alternate_ident_required',
                                             'alternate_ident',
                                             'target_ident_required',
                                             'target_ident',
                                             'allowed_callers_required',
                                             'allowed_callers',
                                             'validate_option_available',
                                             'validate_option',
                                             'validate_fail_ident',
                                             'submit_option_available',
                                             'submit_option',
                                             'submit_required',
                                             'submit_list',
                                             'fail_ident',
                                             'field_options',
                                             'field_values_list',
                                             'field_list',
                                             'single_field_value',
                                             'single_field'
                                            ])


ValidatorInfo = namedtuple('ValidatorInfo', ['validator',
                                             'module_name',
                                             'message',
                                             'message_ref',
                                             'displaywidget',
                                             'allowed_values',
                                             'val_args'
                                            ])









"""A Respond object instance is placed in a Respond page,
   The instance is callable, and the respondpage calls it
   to provide the page action"""

import json, collections

from http import cookies
from string import Template

from .. import skiboot, tag
from ..excepts import ValidateError, ServerError, FailPage, ErrorMessage

from . import Respond


class SubmitData(Respond):
    """
This responder is used in conjunction with previous responders that validates and stores form data in call_data.
Given optional submit_list strings, they will be passed to the user provided submit_data function
in the submit_list argument.
If submit_data raises a FailPage then the fail_ident page will be called.
"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': True,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""

        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)

        try:
            self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, get the target page
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


class ColourSubstitute(Respond):
    """
This responder only applies where the final page returned is a css page.
It will call your submit_data function which should return a dictionary of strings as keys
and colour strings as values.  The keys will be searched for in the target page
with format '${keystring}' and where found the colour value will be placed there.
If submit_data raises a FailPage then the fail page will be called unchanged.
"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""

        try:
            colours = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # return fail page unchanged
            return self.get_fail_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, get the target page
        csspage = self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        if colours and isinstance(colours, dict):
            csspage.colour_substitution = colours
        return csspage



class SetCookies(Respond):
    """
Sets cookies, submit_data should return an instance of http.cookies.BaseCookie
"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data to get a http.cookies.BaseCookie object"""

        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)

        try:
            sendcookies = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
             self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        page = self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # sets the cookies in the page headers
        if sendcookies:
            if not isinstance(sendcookies, cookies.BaseCookie):
                raise ServerError(message = "Returned cookies from submit_data not valid") 
            # set cookies in target_page
            for morsel in sendcookies.values():
                page.headers.append(("Set-Cookie", morsel.OutputString()))
        return page


class GetDictionaryDefaults(Respond):
    """
Web browsers do not send empty fields, therefore a submitted dictionary may have items missing. This responder calls
your submit_data function which should return a dictionary with default values.  Any missing fields
in the form data are then filled in with these defaults. 
The call to submit data will have the 'widgfield':widgfield tuple in the submit dictionary
"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates a list of allowed caller idents is required
    allowed_callers_required = True

    # The widgfield to test
    widgfield_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': True,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted



    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Gets the target page, filling in the form data"
        if caller_page is None:
            raise ValidateError()
        self._check_allowed_callers(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # previous caller is allowed

        skicall.submit_dict['widgfield']=self.widgfield.to_tuple_no_i()
        try:
            # and send the widgfield to submit_data
            defaultdict = self._submit_data( ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        if not isinstance(defaultdict, dict):
            raise ServerError(message = "Returned value from submit_data not valid")

        # if widgfield empty
        if (self.widgfield not in form_data) or (not form_data[self.widgfield]):
            form_data[self.widgfield] = defaultdict
            return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)

        formdict = form_data[self.widgfield]
        if not isinstance(formdict, dict):
            raise ValidateError()

        # check if an unexpected item has been submitted
        for field, val in formdict.items():
            if field not in defaultdict:
                raise ValidateError()

        # fill in any missing key values
        for field, val in defaultdict.items():
            if field not in formdict:
                formdict[field] = val

        # so all ok, get the target page
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


class FieldStoreSubmit(Respond):
    """Takes submitted data from the received form with the given field (regardless of widget name - only uses field name to choose data),
       and stores the data in the dictionary submit_dict under key 'received'.  The dictionary keys in received used to store are the
       widgfield tuple of the submitting widgets.
       Then calls the submit_data function, if no fieldname matches are found submit_data is still called, but no widgfield tuple key
       is inserted into submit_dict"""

    # This indicates a target page ident is required
    target_ident_required = True

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': False,           # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': True}            # A single field is accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):

        # field_name
        field_name = ''
        for key in self.fields:
            field_name = key
        if not field_name:
            raise ValidateError()

        skicall.submit_dict['field'] = field_name

        received = {}

        for field, value in form_data.items():
            # field is a widgfield object
            if field_name == field.f:
               if isinstance(value, list) or isinstance(value, dict):
                   received[field.to_tuple_no_i()] = value.copy()
               else:
                   received[field.to_tuple_no_i()] = value
        skicall.submit_dict['received'] = received

        try:
            self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, get the target page
        return self.get_target_page(skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)


class _JSON(object):
    """An object used by SubmitJSON responder"""

    page_type = "SubmitJSON"

    def __init__(self, ident, jsondict):
        if not jsondict:
            self.jsondict = json.dumps({}).encode('UTF-8')
        else:
            self.jsondict =  json.dumps(jsondict).encode('UTF-8')
        self.status = '200 OK'
        self.headers = [('content-type', 'application/json'),
                        ('cache-control','no-cache, no-store, must-revalidate'),
                        ('Pragma', 'no-cache'),
                        ( 'Expires', '0'),
                        ('content-length', str(len(self.jsondict)))]
        self.ident = ident
        self.ident_data = None
        # Set by end_call
        self.session_cookie = ()
        self.language_cookie = ()

    def import_sections(self, page_data):
        "Only used by Template and SVG, everything else just returns"
        return

    def show_error(self, error_messages=[]):
        return

    def set_values(self, page_data):
        """Checks for header and status values"""
        if not page_data:
            return
        if ('status' in page_data) and page_data['status']:
            self.status = page_data['status']
        if ('headers' in page_data) and page_data['headers']:
            self.headers = page_data['headers']

    def get_status(self):
        "Returns (status, headers)"
        return self.status, self.headers

    def update(self, environ, call_data, lang, ident_list=[]):
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def data(self):
         return [self.jsondict]



class SubmitJSON(Respond):
    """
This responder is intended to provide JSON data as a service to an external application,
submit_data should return a dictionary which will be set into a _JSON object
This dictionary is not widgfields, but is any data you like which can be transformed to JSON.
If a dictionary is successfully returned, the given JSON dictionary will be returned to the client.
Only page_data['status'] and page_data['headers'] will be used if given.
"""

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted

    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""

        try:
            jsondict = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, return JSON file
        return _JSON(ident_list[-1], jsondict)


class _PlainText(object):
    """An object used by SubmitPlainText responder"""

    page_type = "SubmitPlainText"

    def __init__(self, ident, text):
        if not text:
            self.text = "".encode('UTF-8')
        else:
            self.text = text.encode('UTF-8')
        self.status = '200 OK'
        self.headers = [('content-type', 'text/plain'),
                        ('content-length', str(len(self.text)))]
        self.ident = ident
        self.ident_data = None
        # Set by end_call
        self.session_cookie = ()
        self.language_cookie = ()

    def import_sections(self, page_data):
        "Only used by Template and SVG, everything else just returns"
        return

    def show_error(self, error_messages=[]):
        return

    def set_values(self, page_data):
        """Checks for header and status values"""
        if not page_data:
            return
        if ('status' in page_data) and page_data['status']:
            self.status = page_data['status']
        if ('headers' in page_data) and page_data['headers']:
            self.headers = page_data['headers']

    def get_status(self):
        "Returns (status, headers)"
        return self.status, self.headers

    def update(self, environ, call_data, lang, ident_list=[]):
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def data(self):
         return [self.text]



class SubmitPlainText(Respond):
    """
This responder is intended to return plain text to the client.
submit_data should return a string which will be returned to the client.
Only page_data['status'] and page_data['headers'] will be used if given
"""

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""
        try:
            text = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, return a _PlainText instance
        return _PlainText(ident_list[-1], text)


class _CSS(object):
    """An object used by SubmitCSS responder"""

    page_type = "SubmitCSS"

    def __init__(self, ident, style=collections.OrderedDict()):

        # self.style is a dictionary with keys being css selectors
        # and values being a list of two element lists
        # acting as css declaration blocks.
        self.style = style
        self.colour_substitution = {}
        # if self.style_binary given a non empty value, overrides anything else
        self.style_binary = []
        # imports
        self.imports = []
        self.status = '200 OK'
        self.headers = [('content-type', 'text/css'),
                        ('cache-control','no-cache, no-store, must-revalidate'),
                        ('Pragma', 'no-cache'),
                        ('Expires', '0')]
        self.ident = ident
        self.ident_data = None
        # Set by end_call
        self.session_cookie = ()
        self.language_cookie = ()


    # property style is a dictionary of lists
    def get_style(self):
        return self._style

    def set_style(self, style):
        if isinstance(style, collections.OrderedDict):
            self._style = style
        elif isinstance(style, dict):
            self._style = collections.OrderedDict([(selector,value) for selector,value in style.items()])
        else:
            self._style = collections.OrderedDict()

    style = property(get_style, set_style)

    def selector_list(self):
        return list(self._style.keys())

    def selector_properties(self, selector):
        "Returns the property strings of the selector"
        style_text = ""
        if selector not in self._style:
            return ""
        value = self._style[selector]
        if value:
            # value is a list of two element lists
            # created value_list, which is value, sorted by the first element in each sub list
            value_list = sorted(value, key=lambda val : val[0])
            for a,b in value_list:
                style_text += "{a} : {b};\n".format(a=a, b=b)
        return style_text

    def import_sections(self, page_data):
        "Only used by Template and SVG, everything else just returns"
        return

    def show_error(self, error_messages=[]):
        return

    def set_values(self, page_data):
        """Checks for header and status values"""
        if not page_data:
            return
        if '@import' in page_data:
            if isinstance(page_data['@import'], list):
                self.imports = page_data['@import']
            elif isinstance(page_data['@import'], str):
                self.imports = [page_data['@import']]
            del page_data['@import']
        if ('status' in page_data) and page_data['status']:
            self.status = page_data['status']
        if ('headers' in page_data) and page_data['headers']:
            self.headers = page_data['headers']

    def get_status(self):
        "Returns (status, headers)"
        return self.status, self.headers

    def update(self, environ, call_data, lang, ident_list=[]):
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def data(self):
        "Returns the page as a list of binary strings"
        if self.style_binary:
            return self.style_binary
        if not self._style:
            return []
        style_binary = ["@charset \"UTF-8\";\n".encode('UTF-8')]
        if self.imports:
            for imp in self.imports:
                style_binary.append("@import : {imp};\n".format(imp=imp.strip(';')).encode('UTF-8'))
        for selector,value in self._style.items():
            if value:
                style_binary.append("\n{selector} {{".format(selector=selector).encode('UTF-8'))
                # value is a list of two element lists
                # created value_list, which is value, sorted by the first element in each sub list
                value_list = sorted(value, key=lambda val : val[0])
                for a,b in value_list:
                    if self.colour_substitution and ('$' in b):
                        c = Template(b)
                        b= c.safe_substitute(self.colour_substitution)
                    style_binary.append("\n{a} : {b};".format(a=a, b=b).encode('UTF-8'))
                style_binary.append("}\n".encode('UTF-8'))
        return style_binary


class SubmitCSS(Respond):
    """
This responder is intended to return a CSS file to the client.
submit_data should return a style which will be set into a dynamically created
CSS page.
The style is an ordered dictionary with keys being css selectors and values being a
list of two element lists acting as css declaration blocks.
"""

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""

        try:
            styledict = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, return a _CSS instance
        return _CSS(ident_list[-1], styledict)


class MediaQuery(Respond):
    """
Given media queries and CSS page targets, wraps the targets with the media queries
"""

    # The form data can be submitted
    submit_option_available = True

    # Options for the fields argument
    field_options = {'fields': True,                  # If False, no fields are expected
                     'widgfields':False,               # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,        # if fields is True, then False here means fields must be supplied
                     'field_values': True,            # if True, field values are used
                     'empty_values_allowed':False,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        "Matches the value given in field self.widgfield against the fields given"

        media_target =  self.fields.copy()

        # update media target with result of submit_data
        if self.submit_option:
            try:
                skicall.submit_dict['media_target'] = media_target.copy()
                mediadict = self._submit_data(ident_list, skicall)
                if mediadict:
                    media_target.update(mediadict)
            except FailPage as e:
                # raises a PageError exception
                self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)

        if not media_target:
            raise ServerError(message = "No media queries found")

        style_binary = ["@charset \"UTF-8\";\n".encode('UTF-8')]
        
        for media_query, target_ident in media_target.items():
            target = self.get_page_from_ident(target_ident, skicall, form_data, caller_page, ident_list.copy(), proj_ident, rawformdata)
            if hasattr(target, 'style'):
                target_style = target.style
            else:
                continue
            if not target_style:
                continue
            style_binary.append("\n{media_query} {{".format(media_query=media_query).encode('UTF-8'))
            for selector,value in target_style.items():
                if value:
                    style_binary.append("\n{selector} {{".format(selector=selector).encode('UTF-8'))
                    # value is a list of two element lists
                    # created value_list, which is value, sorted by the first element in each sub list
                    value_list = sorted(value, key=lambda val : val[0])
                    for a,b in value_list:
                        if target.colour_substitution and ('$' in b):
                            c = Template(b)
                            b = c.safe_substitute(target.colour_substitution)
                        style_binary.append("\n{a} : {b};".format(a=a, b=b).encode('UTF-8'))
                    style_binary.append("}\n".encode('UTF-8'))
            style_binary.append("}\n".encode('UTF-8'))

        css_object = _CSS(ident_list[-1], {})
        css_object.style_binary = style_binary
        return css_object


class _Iterator(object):
    """An object used by SubmitIterator responder"""

    page_type = "SubmitIterator"

    def __init__(self, ident, biniterator):
        self.biniterator = biniterator
        self.status = '200 OK'
        self.headers = [('cache-control','no-cache, no-store, must-revalidate'),
                        ('Pragma', 'no-cache'),
                        ( 'Expires', '0')]
        self.ident = ident
        self.ident_data = None
        # Set by end_call
        self.session_cookie = ()
        self.language_cookie = ()

    def import_sections(self, page_data):
        "Only used by Template and SVG, everything else just returns"
        return

    def show_error(self, error_messages=[]):
        return

    def set_values(self, page_data):
        """Checks for header, status and mimetype values.  Note; mimetype only used if headers not given"""
        if not page_data:
            return
        if ('status' in page_data) and page_data['status']:
            self.status = page_data['status']
        if ('headers' in page_data) and page_data['headers']:
            self.headers = page_data['headers']
            return
        if ('mimetype' in page_data) and page_data['mimetype']:
            self.headers.append(('content-type', page_data['mimetype']))
        if ('content-length' in page_data) and page_data['content-length']:
            self.headers.append(('content-length', page_data['content-length']))

    def get_status(self):
        "Returns (status, headers)"
        return self.status, self.headers

    def update(self, environ, call_data, lang, ident_list=[]):
        if self.session_cookie:
            self.headers.append(self.session_cookie)
        if self.language_cookie:
            self.headers.append(self.language_cookie)

    def data(self):
         return self.biniterator


class SubmitIterator(Respond):
    """
submit_data should return a binary file iterator
"""

    # This indicates an optional submit_list and fail_ident is required
    submit_required = True

    # Options for the fields argument
    field_options = {'fields': False,                  # If False, no fields are expected
                     'widgfields':False,              # If True, fields are widgfields, if False, can be other constants
                     'widgfield_values':False,        # If True the field values are widgfields
                     'fields_optional': False,         # if fields is True, then False here means fields must be supplied
                     'field_values':False,            # if True, field values are used
                     'empty_values_allowed':True,     # If True, '' is a valid value, if False, some data must be provided
                     'single_field': False}           # Multiple fields accepted


    def _respond(self, skicall, form_data, caller_page, ident_list, proj_ident, rawformdata):
        """Calls submit_data"""
        try:
            biniterator = self._submit_data(ident_list, skicall)
        except FailPage as e:
            # raises a PageError exception
            self.raise_error_page([e.errormessage], skicall, form_data, caller_page, ident_list, proj_ident, rawformdata)
        # so all ok, return a _Iterator instance
        return _Iterator(ident_list[-1], biniterator)







import os, random, collections

from skipole import skilift
from skipole.skilift import off_piste, fromjson, editpage

from .. import utils, css_styles
from skipole import FailPage, ValidateError, ServerError, GoTo


def retrieve_index_data(skicall):
    "Retrieves all field data for admin index page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data)

    project = call_data['editedprojname']

    projectinfo = skilift.project_info(project)

    page_data["projversion", "input_text"] = projectinfo.version
    page_data["brief:input_text"] = projectinfo.brief
    page_data["brief:bottomtext"] = "Current value: " + projectinfo.brief
    page_data["deflang:input_text"] = projectinfo.default_language
    page_data["deflang:bottomtext"] = "Current value: " + projectinfo.default_language

    if "root_path" in call_data:
        # "root_path" is set in call_data, so return it
        page_data["rtpath:input_text"] = call_data["root_path"]
    else:
        # "root_path" not in call_data, so return the current root_path
        page_data["rtpath:input_text"] = projectinfo.path

    if skilift.get_debug():
        page_data["debugstatus", "para_text"] = "Debug mode is ON"
        page_data["debugtoggle", "button_text"] = "Set Debug OFF"
    else:
        page_data["debugstatus", "para_text"] = "Debug mode is OFF"
        page_data["debugtoggle", "button_text"] = "Set Debug ON"

    # create table of projects
    subprojects = projectinfo.subprojects
    col1 = []
    col2 = []
    col3 = []
    for proj, suburl in subprojects.items():
        projinfo = skilift.project_info(proj)
        col1.append(proj)
        col2.append(suburl)
        col3.append(projinfo.brief)
    # append the final row showing this edited project
    col1.append(project)
    col2.append(projectinfo.path)
    col3.append(projectinfo.brief)
    page_data['subs','col1'] = col1
    page_data['subs','col2'] = col2
    page_data['subs','col3'] = col3


    ctable = []
    for proj, suburl in subprojects.items():
        projinfo = skilift.project_info(proj)
        ctable.append([proj, suburl, projinfo.brief, proj, '', proj, '', True, True])
    # append the final row showing this edited project
    ctable.append([project, projectinfo.path, projectinfo.brief, '', '', '', '', False, False])
    page_data["projtable:contents"] = ctable

    # list directories under projects_dir()
    page_data['sdd1:option_list'] = []

    # sdd1 must be deleted
    page_data['sdd1:show_add_project'] = False
    page_data['sdd1:selectvalue'] = ''


def _clear_index_input_accepted(page_data):
    "Used by JSON to remove set_input_accepted flags from input fields"
    page_data['deflang','set_input_accepted'] = False
    page_data['projversion','set_input_accepted'] = False
    page_data['rtpath','set_input_accepted'] = False
    page_data['brief','set_input_accepted'] = False


def retrieve_help(skicall):
    "Uses skicall.textblock_text to get text help for the admin pages"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # clears any session data
    utils.clear_call_data(call_data)
    page_data.clear()
    caller_ident = call_data['caller_ident']
    if not caller_ident:
        return
    textref = 'page.' + str(caller_ident[1])
    text = skicall.textblock(textref)
    if not text:
        text = "No help text for page %s has been found" % caller_ident[1]
    page_data[("adminhead","show_help","para_text")] = "\n" + text
    page_data[("adminhead","show_help","hide")] = False


def get_text(skicall):
    """Finds any widget submitting 'get_field' with value of a textblock ref, returns
       page_data with key widget with field 'div_content' and value the textblock text"""

    page_data = skicall.page_data

    if 'received' not in skicall.submit_dict:
        return
    received_widgfields = skicall.submit_dict['received']
    for key, val in received_widgfields.items():
        if isinstance(key, tuple) and (key[-1] == 'get_field'):
            text = skicall.textblock(val)
            if text is None:
                continue
            text = text.replace('\n', '\n<br />')
            if len(key) == 3:
                page_data[(key[0], key[1],'div_content')] = text
                page_data[(key[0], key[1],'hide')] = False
            elif len(key) == 2:
                page_data[(key[0],'div_content')] = text
                page_data[(key[0],'hide')] = False


def retrieve_colour_data(skicall):
    "Retrieves all field data for admin colour page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # get default admin background color from project data
    adminbackcol = skicall.proj_data['adminbackcol']
    # get individual admin colors from project data
    colours = skicall.proj_data['colours']

    page_data["colhextest", "input_text"] = adminbackcol
    admincol = css_styles.hex_int(adminbackcol)
    page_data["red","input_text"] = str(admincol[0])
    page_data["green","input_text"] = str(admincol[1])
    page_data["blue","input_text"] = str(admincol[2])
    admin_h,admin_s,admin_l = css_styles.rgb_hsl(*admincol)
    page_data["hue","input_text"] = str(int(admin_h*360))
    page_data["saturation","input_text"] = str(int(admin_s*100))
    page_data["lightness","input_text"] = str(int(admin_l*100))



    l5_color = colours['w3_theme_l5_color']
    l5_background_color = colours['w3_theme_l5_background_color']
    l4_color = colours['w3_theme_l4_color']
    l4_background_color = colours['w3_theme_l4_background_color']
    l3_color = colours['w3_theme_l3_color']
    l3_background_color = colours['w3_theme_l3_background_color']
    l2_color = colours['w3_theme_l2_color']
    l2_background_color = colours['w3_theme_l2_background_color']
    l1_color = colours['w3_theme_l1_color']
    l1_background_color = colours['w3_theme_l1_background_color']
    d1_color = colours['w3_theme_d1_color']
    d1_background_color = colours['w3_theme_d1_background_color']
    d2_color = colours['w3_theme_d2_color']
    d2_background_color = colours['w3_theme_d2_background_color']
    d3_color = colours['w3_theme_d3_color']
    d3_background_color = colours['w3_theme_d3_background_color']
    d4_color = colours['w3_theme_d4_color']
    d4_background_color = colours['w3_theme_d4_background_color']
    d5_color = colours['w3_theme_d5_color']
    d5_background_color = colours['w3_theme_d5_background_color']
    light_color = colours['w3_theme_light_color']
    light_background_color = colours['w3_theme_light_background_color']
    dark_color = colours['w3_theme_dark_color']
    dark_background_color = colours['w3_theme_dark_background_color']
    action_color = colours['w3_theme_action_color']
    action_background_color = colours['w3_theme_action_background_color']
    theme_color = colours['w3_theme_color']
    theme_background_color = colours['w3_theme_background_color']
    text_color = colours['w3_text_theme_color']
    border_color = colours['w3_border_theme_color']
    hover_color = colours['w3_hover_theme_hover_color']
    hover_background_color = colours['w3_hover_theme_hover_background_color']
    hover_text_color = colours['w3_hover_text_theme_color']
    hover_border_color = colours['w3_hover_border_theme_hover_color']


    # l5_color
    page_data["l5_color","inputcol","label"] = "w3-theme-l5 color:"
    page_data["l5_color","inputcol","input_text"] = l5_color
    page_data["l5_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l5_color,l5_background_color)
    page_data["l5_color", "boxcol","set_html"] = l5_color

    # l5_background_color
    page_data["l5_background_color","inputcol","label"] = "w3-theme-l5 background color:"
    page_data["l5_background_color","inputcol","input_text"] = l5_background_color
    page_data["l5_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l5_background_color, l5_color)
    page_data["l5_background_color", "boxcol","set_html"] = l5_background_color

    # l4_color
    page_data["l4_color","inputcol","label"] = "w3-theme-l4 color:"
    page_data["l4_color","inputcol","input_text"] = l4_color
    page_data["l4_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l4_color,l4_background_color)
    page_data["l4_color", "boxcol","set_html"] = l4_color

    # l4_background_color
    page_data["l4_background_color","inputcol","label"] = "w3-theme-l4 background color:"
    page_data["l4_background_color","inputcol","input_text"] = l4_background_color
    page_data["l4_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l4_background_color,l4_color)
    page_data["l4_background_color", "boxcol","set_html"] = l4_background_color

    # l3_color
    page_data["l3_color","inputcol","label"] = "w3-theme-l3 color:"
    page_data["l3_color","inputcol","input_text"] = l3_color
    page_data["l3_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l3_color,l3_background_color)
    page_data["l3_color", "boxcol","set_html"] = l3_color

    # l3_background_color
    page_data["l3_background_color","inputcol","label"] = "w3-theme-l3 background color:"
    page_data["l3_background_color","inputcol","input_text"] = l3_background_color
    page_data["l3_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l3_background_color,l3_color)
    page_data["l3_background_color", "boxcol","set_html"] = l3_background_color

    # l2_color
    page_data["l2_color","inputcol","label"] = "w3-theme-l2 color:"
    page_data["l2_color","inputcol","input_text"] = l2_color
    page_data["l2_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l2_color,l2_background_color)
    page_data["l2_color", "boxcol","set_html"] = l2_color

    # l2_background_color
    page_data["l2_background_color","inputcol","label"] = "w3-theme-l2 background color:"
    page_data["l2_background_color","inputcol","input_text"] = l2_background_color
    page_data["l2_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l2_background_color,l2_color)
    page_data["l2_background_color", "boxcol","set_html"] = l2_background_color

    # l1_color
    page_data["l1_color","inputcol","label"] = "w3-theme-l1 color:"
    page_data["l1_color","inputcol","input_text"] = l1_color
    page_data["l1_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l1_color,l1_background_color)
    page_data["l1_color", "boxcol","set_html"] = l1_color

    # l1_background_color
    page_data["l1_background_color","inputcol","label"] = "w3-theme-l1 background color:"
    page_data["l1_background_color","inputcol","input_text"] = l1_background_color
    page_data["l1_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (l1_background_color,l1_color)
    page_data["l1_background_color", "boxcol","set_html"] = l1_background_color

    # d1_color
    page_data["d1_color","inputcol","label"] = "w3-theme-d1 color:"
    page_data["d1_color","inputcol","input_text"] = d1_color
    page_data["d1_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d1_color,d1_background_color)
    page_data["d1_color", "boxcol","set_html"] = d1_color

    # d1_background_color
    page_data["d1_background_color","inputcol","label"] = "w3-theme-d1 background color:"
    page_data["d1_background_color","inputcol","input_text"] = d1_background_color
    page_data["d1_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d1_background_color,d1_color)
    page_data["d1_background_color", "boxcol","set_html"] = d1_background_color

    # d2_color
    page_data["d2_color","inputcol","label"] = "w3-theme-d2 color:"
    page_data["d2_color","inputcol","input_text"] = d2_color
    page_data["d2_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d2_color,d2_background_color)
    page_data["d2_color", "boxcol","set_html"] = d2_color

    # d2_background_color
    page_data["d2_background_color","inputcol","label"] = "w3-theme-d2 background color:"
    page_data["d2_background_color","inputcol","input_text"] = d2_background_color
    page_data["d2_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d2_background_color,d2_color)
    page_data["d2_background_color", "boxcol","set_html"] = d2_background_color

    # d3_color
    page_data["d3_color","inputcol","label"] = "w3-theme-d3 color:"
    page_data["d3_color","inputcol","input_text"] = d3_color
    page_data["d3_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d3_color,d3_background_color)
    page_data["d3_color", "boxcol","set_html"] = d3_color

    # d3_background_color
    page_data["d3_background_color","inputcol","label"] = "w3-theme-d3 background color:"
    page_data["d3_background_color","inputcol","input_text"] = d3_background_color
    page_data["d3_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d3_background_color,d3_color)
    page_data["d3_background_color", "boxcol","set_html"] = d3_background_color

    # d4_color
    page_data["d4_color","inputcol","label"] = "w3-theme-d4 color:"
    page_data["d4_color","inputcol","input_text"] = d4_color
    page_data["d4_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d4_color,d4_background_color)
    page_data["d4_color", "boxcol","set_html"] = d4_color

    # d4_background_color
    page_data["d4_background_color","inputcol","label"] = "w3-theme-d4 background color:"
    page_data["d4_background_color","inputcol","input_text"] = d4_background_color
    page_data["d4_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d4_background_color,d4_color)
    page_data["d4_background_color", "boxcol","set_html"] = d4_background_color

    # d5_color
    page_data["d5_color","inputcol","label"] = "w3-theme-d5 color:"
    page_data["d5_color","inputcol","input_text"] = d5_color
    page_data["d5_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d5_color,d5_background_color)
    page_data["d5_color", "boxcol","set_html"] = d5_color

    # d5_background_color
    page_data["d5_background_color","inputcol","label"] = "w3-theme-d5 background color:"
    page_data["d5_background_color","inputcol","input_text"] = d5_background_color
    page_data["d5_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (d5_background_color,d5_color)
    page_data["d5_background_color", "boxcol","set_html"] = d5_background_color


    # light_color
    page_data["light_color","inputcol","label"] = "w3-theme-light color:"
    page_data["light_color","inputcol","input_text"] = light_color
    page_data["light_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (light_color,light_background_color)
    page_data["light_color", "boxcol","set_html"] = light_color

    # light_background_color
    page_data["light_background_color","inputcol","label"] = "w3-theme-light background color:"
    page_data["light_background_color","inputcol","input_text"] = light_background_color
    page_data["light_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (light_background_color,light_color)
    page_data["light_background_color", "boxcol","set_html"] = light_background_color

    # dark_color
    page_data["dark_color","inputcol","label"] = "w3-theme-dark color:"
    page_data["dark_color","inputcol","input_text"] = dark_color
    page_data["dark_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (dark_color,dark_background_color)
    page_data["dark_color", "boxcol","set_html"] = dark_color

    # dark_background_color
    page_data["dark_background_color","inputcol","label"] = "w3-theme-dark background color:"
    page_data["dark_background_color","inputcol","input_text"] = dark_background_color
    page_data["dark_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (dark_background_color,dark_color)
    page_data["dark_background_color", "boxcol","set_html"] = dark_background_color

    # action_color
    page_data["action_color","inputcol","label"] = "w3-theme-action color:"
    page_data["action_color","inputcol","input_text"] = action_color
    page_data["action_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (action_color,action_background_color)
    page_data["action_color", "boxcol","set_html"] = action_color

    # action_background_color
    page_data["action_background_color","inputcol","label"] = "w3-theme-action background color:"
    page_data["action_background_color","inputcol","input_text"] = action_background_color
    page_data["action_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (action_background_color,action_color)
    page_data["action_background_color", "boxcol","set_html"] = action_background_color

    # theme_color
    page_data["theme_color","inputcol","label"] = "w3-theme color:"
    page_data["theme_color","inputcol","input_text"] = theme_color
    page_data["theme_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (theme_color,theme_background_color)
    page_data["theme_color", "boxcol","set_html"] = theme_color

    # theme_background_color
    page_data["theme_background_color","inputcol","label"] = "w3-theme background color:"
    page_data["theme_background_color","inputcol","input_text"] = theme_background_color
    page_data["theme_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (theme_background_color,theme_color)
    page_data["theme_background_color", "boxcol","set_html"] = theme_background_color

    # text_color
    page_data["text_color","inputcol","label"] = "w3-text-theme color:"
    page_data["text_color","inputcol","input_text"] = text_color
    page_data["text_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (text_color,theme_background_color)
    page_data["text_color", "boxcol","set_html"] = text_color

    # border_color
    page_data["border_color","inputcol","label"] = "w3-border-theme color:"
    page_data["border_color","inputcol","input_text"] = border_color
    page_data["border_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (border_color,theme_background_color)
    page_data["border_color", "boxcol","set_html"] = border_color

    # hover_color
    page_data["hover_color","inputcol","label"] = "w3-hover-theme:hover color:"
    page_data["hover_color","inputcol","input_text"] = hover_color
    page_data["hover_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (hover_color,hover_background_color)
    page_data["hover_color", "boxcol","set_html"] = hover_color

    # hover_background_color
    page_data["hover_background_color","inputcol","label"] = "w3-hover-theme:hover background color:"
    page_data["hover_background_color","inputcol","input_text"] = hover_background_color
    page_data["hover_background_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (hover_background_color,hover_color)
    page_data["hover_background_color", "boxcol","set_html"] = hover_background_color

    # hover_text_color
    page_data["hover_text_color","inputcol","label"] = "w3-hover-text-theme:hover color:"
    page_data["hover_text_color","inputcol","input_text"] = hover_text_color
    page_data["hover_text_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (hover_text_color,hover_background_color)
    page_data["hover_text_color", "boxcol","set_html"] = hover_text_color

    # hover_border_color
    page_data["hover_border_color","inputcol","label"] = "w3-hover-border-theme:hover color:"
    page_data["hover_border_color","inputcol","input_text"] = hover_border_color
    page_data["hover_border_color", "boxcol", "style"] = "float: right; width: 20em; background-color: %s; color: %s; border: solid white 2px; padding: 5px;" % (hover_border_color,hover_background_color)
    page_data["hover_border_color", "boxcol","set_html"] = hover_border_color


def goto_edit_item(skicall):
    "Goes to the edit page, folder etc depending on the item submitted"

    call_data = skicall.call_data
    page_data = skicall.page_data

    edited_item = call_data["edited_item"]
    project = call_data['editedprojname']

    # clear call data
    utils.clear_call_data(call_data)

    itemnumber = skilift.get_itemnumber(project, edited_item)
    if itemnumber is None:
        raise FailPage(message="Item not found")

    item = skilift.item_info(project, itemnumber)
    if item is None:
        raise FailPage(message="Item not found")

    if item.item_type == "Folder":
        call_data['folder_number'] = itemnumber
        call_data['fchange'] = item.change
        raise GoTo(target=22008, clear_submitted=True)

    call_data['page_number'] = itemnumber
    call_data['pchange'] = item.change

    if item.item_type == "TemplatePage":
        raise GoTo(target=23207, clear_submitted=True)
    elif item.item_type == "RespondPage":
        raise GoTo(target=26007, clear_submitted=True)
    elif item.item_type == "FilePage":
        raise GoTo(target=29007, clear_submitted=True)
    elif item.item_type == "CSS":
        raise GoTo(target=28007, clear_submitted=True)
    elif item.item_type == "JSON":
        raise GoTo(target=20407, clear_submitted=True)
    elif item.item_type == "SVG":
        raise GoTo(target=23407, clear_submitted=True)

    del call_data['page_number']
    del call_data['pchange']
    raise FailPage(message="Item not found")
    

def submit_language(skicall):
    "Sets the default language of the edited project"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if "default_language" not in call_data:
        raise FailPage(message = 'Invalid call')
    try:
        skilift.set_proj_default_language(project, call_data["default_language"])
    except ServerError as e:
        raise FailPage(message = e.message)
    _clear_index_input_accepted(page_data)
    page_data['deflang','set_input_accepted'] = True
    call_data['status'] = 'Language set'


def submit_brief(skicall):
    "Sets the brief description of the edited project"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if ('brief','input_text') not in call_data:
        raise FailPage(message='Invalid call')
    try:
        skilift.set_proj_brief(project, call_data['brief','input_text'])
    except ServerError as e:
        raise FailPage(message = e.message)
    call_data['editedprojbrief'] = call_data['brief','input_text']
    page_data['brief','set_input_accepted'] = True
    call_data['status'] = 'Project brief set'


def set_version(skicall):
    "Sets the version of the edited project"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if ('projversion','input_text') not in call_data:
        raise FailPage(message='Invalid call')
    try:
        skilift.set_proj_version(project, call_data['projversion','input_text'])
    except ServerError as e:
        raise FailPage(message = e.message)
    _clear_index_input_accepted(page_data)
    call_data['editedprojversion'] = call_data['projversion','input_text']
    page_data['projversion','set_input_accepted'] = True
    page_data["adminhead","page_head","large_text"] = "Project: %s version: %s" % (project,call_data['projversion','input_text'])
    call_data['status'] = 'Project version set'


def debugtoggle(skicall):

    call_data = skicall.call_data
    page_data = skicall.page_data

    if skilift.get_debug():
        skilift.set_debug(False)
        page_data["debugtoggle", "button_text"] = "Set Debug ON"
        call_data['status'] = 'Debug mode set OFF'
    else:
        skilift.set_debug(True)
        page_data["debugtoggle", "button_text"] = "Set Debug OFF"
        call_data['status'] = 'Debug mode set ON'


def _submit_saveproject(skicall):
    "save the project textblock and project.json files"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    projinfo = skilift.project_info(project)
    # Create textblock json files
    accesstextblocks = skilift.get_accesstextblocks(project)
    accesstextblocks.save()
    # create project.json
    fromjson.project_to_json(project)


def json_submit_saveproject(skicall):
    "save the project textblock and project.json files"

    call_data = skicall.call_data
    page_data = skicall.page_data

    _submit_saveproject(skicall)
    page_data[("saveresult","para_text")] = "Project data saved to JSON files"
    page_data[("saveresult","show_para")] = True
    # clears any session data
    utils.clear_call_data(call_data)
    call_data['status'] = "Project data saved to JSON files"


def html_submit_saveproject(skicall):
    "save the project textblock and project.json files"
    call_data = skicall.call_data
    page_data = skicall.page_data

    _submit_saveproject(skicall)
    # clears any session data
    utils.clear_call_data(call_data)
    call_data['status'] = "Project data saved to JSON files"


def submit_hex_color(skicall):
    "set palette background color - from hex"

    call_data = skicall.call_data
    page_data = skicall.page_data

    adminproj = call_data['adminproj']
    palette = css_styles.hex_int(call_data['hexcol'].lower())
    adminbackcol = css_styles.int_hex(*palette)
    # generate colours
    colours = css_styles.get_colours(*palette)
    # set test colours in proj_data
    skilift.set_proj_data(adminproj, 'adminbackcol', adminbackcol)
    skilift.set_proj_data(adminproj, 'colours', colours)
    # set test colours in defaults.json
    fromjson.set_defaults(adminproj, key="backcol", value=adminbackcol)
    ordered_colours = collections.OrderedDict(sorted(colours.items(), key=lambda t: t[0]))
    fromjson.set_defaults(adminproj, key="colours", value=ordered_colours)
    # change name of 220 to avoid css cache
    pchange = editpage.pagechange(adminproj, 220)
    if pchange is None:
        raise ServerError(message="Admin page 220, label w3-theme-ski has not been found")
    newname = "w3-theme-ski-" + str(random.randint(10000, 99999)) + ".css"
    editpage.rename_page(adminproj, 220, pchange, newname)


def set_colour(skicall):
    "sets individual colour string"

    call_data = skicall.call_data
    page_data = skicall.page_data

    adminproj = call_data['adminproj']
    # get colors from project data
    colours = skilift.get_proj_data(adminproj, 'colours')
    if call_data["l5_color","inputcol","input_text"]:
        colours['w3_theme_l5_color'] = call_data["l5_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l5 color set'
    elif call_data["l5_background_color","inputcol","input_text"]:
        colours['w3_theme_l5_background_color'] = call_data["l5_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l5 background color set'
    elif call_data["l4_color","inputcol","input_text"]:
        colours['w3_theme_l4_color'] = call_data["l4_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l4 color set'
    elif call_data["l4_background_color","inputcol","input_text"]:
        colours['w3_theme_l4_background_color'] = call_data["l4_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l4 background color set'
    elif call_data["l3_color","inputcol","input_text"]:
        colours['w3_theme_l3_color'] = call_data["l3_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l3 color set'
    elif call_data["l3_background_color","inputcol","input_text"]:
        colours['w3_theme_l3_background_color'] = call_data["l3_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l3 background color set'
    elif call_data["l2_color","inputcol","input_text"]:
        colours['w3_theme_l2_color'] = call_data["l2_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l2 color set'
    elif call_data["l2_background_color","inputcol","input_text"]:
        colours['w3_theme_l2_background_color'] = call_data["l2_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l2 background color set'
    elif call_data["l1_color","inputcol","input_text"]:
        colours['w3_theme_l1_color'] = call_data["l1_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l1 color set'
    elif call_data["l1_background_color","inputcol","input_text"]:
        colours['w3_theme_l1_background_color'] = call_data["l1_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-l1 background color set'
    elif call_data["d5_color","inputcol","input_text"]:
        colours['w3_theme_d5_color'] = call_data["d5_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d5 color set'
    elif call_data["d5_background_color","inputcol","input_text"]:
        colours['w3_theme_d5_background_color'] = call_data["d5_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d5 background color set'
    elif call_data["d4_color","inputcol","input_text"]:
        colours['w3_theme_d4_color'] = call_data["d4_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d4 color set'
    elif call_data["d4_background_color","inputcol","input_text"]:
        colours['w3_theme_d4_background_color'] = call_data["d4_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d4 background color set'
    elif call_data["d3_color","inputcol","input_text"]:
        colours['w3_theme_d3_color'] = call_data["d3_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d3 color set'
    elif call_data["d3_background_color","inputcol","input_text"]:
        colours['w3_theme_d3_background_color'] = call_data["d3_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d3 background color set'
    elif call_data["d2_color","inputcol","input_text"]:
        colours['w3_theme_d2_color'] = call_data["d2_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d2 color set'
    elif call_data["d2_background_color","inputcol","input_text"]:
        colours['w3_theme_d2_background_color'] = call_data["d2_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d2 background color set'
    elif call_data["d1_color","inputcol","input_text"]:
        colours['w3_theme_d1_color'] = call_data["d1_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d1 color set'
    elif call_data["d1_background_color","inputcol","input_text"]:
        colours['w3_theme_d1_background_color'] = call_data["d1_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-d1 background color set'
    elif call_data["light_color","inputcol","input_text"]:
        colours['w3_theme_light_color'] = call_data["light_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-light color set'
    elif call_data["light_background_color","inputcol","input_text"]:
        colours['w3_theme_light_background_color'] = call_data["light_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-light background color set'
    elif call_data["dark_color","inputcol","input_text"]:
        colours['w3_theme_dark_color'] = call_data["dark_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-dark color set'
    elif call_data["dark_background_color","inputcol","input_text"]:
        colours['w3_theme_dark_background_color'] = call_data["dark_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-dark background color set'
    elif call_data["action_color","inputcol","input_text"]:
        colours['w3_theme_action_color'] = call_data["action_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-action color set'
    elif call_data["action_background_color","inputcol","input_text"]:
        colours['w3_theme_action_background_color'] = call_data["action_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme-action background color set'
    elif call_data["theme_color","inputcol","input_text"]:
        colours['w3_theme_color'] = call_data["theme_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme color set'
    elif call_data["theme_background_color","inputcol","input_text"]:
        colours['w3_theme_background_color'] = call_data["theme_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-theme background color set'
    elif call_data["text_color","inputcol","input_text"]:
        colours['w3_text_theme_color'] = call_data["text_color","inputcol","input_text"]
        call_data['status'] = 'w3-text-theme color set'
    elif call_data["border_color","inputcol","input_text"]:
        colours['w3_border_theme_color'] = call_data["boder_color","inputcol","input_text"]
        call_data['status'] = 'w3-border-theme color set'
    elif call_data["hover_color","inputcol","input_text"]:
        colours['w3_hover_theme_hover_color'] = call_data["hover_color","inputcol","input_text"]
        call_data['status'] = 'w3-hover-theme:hover color set'
    elif call_data["hover_background_color","inputcol","input_text"]:
        colours['w3_hover_theme_hover_background_color'] = call_data["hover_background_color","inputcol","input_text"]
        call_data['status'] = 'w3-hover-theme:hover background color set'
    elif call_data["hover_text_color","inputcol","input_text"]:
        colours['w3_hover_text_theme_color'] = call_data["hover_text_color","inputcol","input_text"]
        call_data['status'] = 'w3-hover-text-theme:hover color set'
    elif call_data["hover_border_color","inputcol","input_text"]:
        colours['w3_hover_border_theme_hover_color'] = call_data["hover_border_color","inputcol","input_text"]
        call_data['status'] = 'w3-hover-border-theme:hover color set'
    else:
        # no colour recognised
        return
    # etc for each colour
    # Then set colours into project data
    skilift.set_proj_data(adminproj, 'colours', colours)
    # set test colours in defaults.json
    fromjson.set_defaults(adminproj, key="colours", value=colours)
    # change name of 220 to avoid css cache
    pchange = editpage.pagechange(adminproj, 220)
    if pchange is None:
        raise ServerError(message="Admin page 220, label w3-theme-ski has not been found")
    newname = "w3-theme-ski-" + str(random.randint(10000, 99999)) + ".css"
    editpage.rename_page(adminproj, 220, pchange, newname)


def ski_theme(skicall):
    "set colours into the w3-theme-ski.css page"

    call_data = skicall.call_data
    page_data = skicall.page_data

    adminproject = call_data['adminproj']
    colours = skilift.get_proj_data(adminproject, 'colours')
    if not colours:
        return {}
    return colours


def set_widgets_css(skicall):
    "sets default css classes into widgets"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data["editedprojname"]
    off_piste.set_widget_css_to_default(project)
    call_data['status'] = 'Widget CSS classes set'









"""
This module defines base exceptions. It can be imported
by any other module as it has no dependencies
"""

class ErrorMessage(object):
    """Acts as a store for attributes message, section, widget

       The message can be either a string, or an empty string

       The following are all optional

       section is the section name (if it is in a section) of the widget where the message is to be displayed
       widget is the widget name where the message is to be displayed
"""
    def __init__(self, message='', section='', widget=''):
        self.message = message
        self.section = section
        self.widget = widget

    def __str__(self):
        return """message: %s
section: %s
widget: %s
code: %s
""" % (self.message,
       self.section,
       self.widget)


class SkiError(Exception):
    """Exception used as a parent to other exception classes that carry message information.
    """

    def __init__(self, message = '', section='', widget=''):
        if message:
            Exception.__init__(self, message)
        self.errormessage = ErrorMessage(message, section, widget)
        # These attributes are filled in by the framework when an error is raised
        self.ident_list = []
        self.form_data = ''

    def get_message(self):
        return self.errormessage.message

    def set_message(self, message):
        self.errormessage.message = message

    message = property(get_message, set_message)

    def get_section(self):
        return self.errormessage.section

    def set_section(self, section):
        self.errormessage.section = section

    section = property(get_section, set_section)

    def get_widget(self):
        return self.errormessage.widget

    def set_widget(self, widget):
        self.errormessage.widget = widget

    widget = property(get_widget, set_widget)


class FailPage(SkiError):
    """Call a responder fail ident page
    """


class ServerError(SkiError):
    """Flag a server problem"""
    def __init__(self, message = '', section='', widget='', status='500 Internal Server Error', code=0):
        SkiError.__init__(self, message, section, widget)
        self.status = status
        # code 0 is default
        # codes 9000 to 9999 are reserved for system use
        self.code = code


class ValidateError(SkiError):
    """Flag a validation error."""
    def __init__(self, message = '', section='', widget='', status='400 Bad Request'):
        SkiError.__init__(self, message, section, widget)
        self.status = status


class PageError(Exception):
    """Exception used to flag a page error - used internally by responders."""

    def __init__(self, page, e_list=[]):
        "Sets an exception with a page"
        self.page = page
        self.e_list = e_list


class GoTo(Exception):
    """Exception used to force a jump to another page

       target being the ident, label or external
                      url of the page to jump to. If a url is given it should start
                      with 'http://' or 'https://' or '//'

       clear_submitted - if True, any submitted data widgets/fields will be cleared,
       clear_page_data - if True, page_data dictionary will be cleared,
       clear_errors - if a FailPage raises an error and the failpage calls
                      a submit function with a GoTo, then setting clear_errors True
                      will remove the error condition from the call, which will therefore
                      not be displayed on the final template page returned

       Note, 'call_data' will not be changed.
"""

    def __init__(self, target, clear_submitted=False, clear_page_data=False, clear_errors=False):
        Exception.__init__(self, "GoTo jump to page " + str(target))
        self.clear_submitted = clear_submitted
        self.clear_page_data = clear_page_data
        self.clear_errors = clear_errors
        self.target = target
        # this is set by the calling responder
        self.proj_ident=None
        # this is set if a FailPage call leads to a responder which raises a GoTo
        self.e_list = []






"Functions to insert an html tage"

from skipole import FailPage, ValidateError, GoTo, ServerError

from skipole.skilift import item_info, editpage, editsection


def retrieve_insert(skicall):
    "Fills in the insert an html element page header"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    if 'page_number' in call_data:
        pageinfo = item_info(project, call_data['page_number'])
        if (pageinfo.item_type != "TemplatePage") and (pageinfo.item_type != "SVG"):
            raise FailPage("Invalid page")
        # Fill in header
        page_data[("adminhead","page_head","large_text")] = "Insert an HTML element into %s,%s" % (project, call_data['page_number'])
    elif 'section_name' in call_data:
        page_data[("adminhead","page_head","large_text")] = "Insert an HTML element into section " + call_data['section_name']
    else:
        raise FailPage("Insertion page/section for html element is missing")


def create_insert(skicall):
    "Creates the html element"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    try:
        part_top, container, location_integers = call_data['location']
        if call_data['newopenclosed'] == 'open':
            opentag = True
        else:
            opentag = False

        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
            page_info = item_info(project, pagenumber)
            if page_info is None:
                raise FailPage("Page to edit not identified")

            if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
                raise FailPage("Page not identified")

            # page to go back to
            target = None
            if part_top == 'head':
                target = "page_head"
            elif part_top == 'body':
                target = "page_body"
            elif part_top == 'svg':
                target = "page_svg"

            if container is not None:
                target = "back_to_container"

            if target is None:
                raise FailPage("Invalid location")

            call_data['pchange'] = editpage.create_html_element_in_page(project,
                                                                        pagenumber,
                                                                        call_data['pchange'],
                                                                        call_data['location'],
                                                                        call_data['newpartname'],
                                                                        call_data['newbrief'],
                                                                        opentag)

        elif 'section_name' in call_data:
            section_name = call_data['section_name']

            if container is not None:
                target = "back_to_container"
            else:
                target = "back_to_section"

            call_data['schange'] = editsection.create_html_element_in_section(project,
                                                                              section_name,
                                                                              call_data['schange'],
                                                                              call_data['location'],
                                                                              call_data['newpartname'],
                                                                              call_data['newbrief'],
                                                                              opentag)

        else:
            raise FailPage("Either a page or section must be specified")

    except ServerError as e:
        raise FailPage(e.message)
    call_data['status'] = 'New tag inserted'
    raise GoTo(target = target, clear_submitted=True)


def file_new_part(skicall):
    "Create new part from uploaded file"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']
    pagenumber = None
    section_name = None

    try:
        part_top, container, location_integers = call_data['location']

        if 'page_number' in call_data:
            pagenumber = call_data['page_number']
            page_info = item_info(project, pagenumber)
            if page_info is None:
                raise FailPage("Page to edit not identified")

            if (page_info.item_type != "TemplatePage") and (page_info.item_type != "SVG"):
                raise FailPage("Page not identified")

            # page to go back to
            target = None
            if part_top == 'head':
                target = "page_head"
            elif part_top == 'body':
                target = "page_body"
            elif part_top == 'svg':
                target = "page_svg"

            if container is not None:
                target = "back_to_container"

            if target is None:
                raise FailPage("Invalid location")

        elif 'section_name' in call_data:
            section_name = call_data['section_name']

            if container is not None:
                target = "back_to_container"
            else:
                target = "back_to_section"

        else:
            raise FailPage("Either a page or section must be specified")

        # get file contents
        file_contents = call_data["uploadpart", "action"]
        json_string = file_contents.decode(encoding='utf-8')
        # create the part
        if pagenumber:
            call_data['pchange'] = editpage.create_part_in_page(project, pagenumber, call_data['pchange'], call_data['location'], json_string)
        else:
            call_data['schange'] = editsection.create_part_in_section(project, section_name, call_data['schange'], call_data['location'], json_string)
    except ServerError as e:
        raise FailPage(message = e.message)

    call_data['status'] = 'New block created'
    raise GoTo(target = target, clear_submitted=True)







"Functions implementing admin page editing"


import re

from skipole import skilift
from skipole.skilift import editwidget

from skipole import ServerError, FailPage, ValidateError, GoTo

# a search for anything none-alphanumeric and not an underscore
_AN = re.compile('[^\w]')


def retrieve_module_list(skicall):
    "this call is to retrieve data for listing widget modules"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # Fill in header
    page_data[("adminhead","page_head","large_text")] = "Choose module"

    # as this page chooses a module, clear any previous chosen module and widget class
    if 'module' in call_data:
        del call_data['module']
    if 'widgetclass' in call_data:
        del call_data['widgetclass']

    # table of widget modules

    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the get field of the link
    # col 3 is the reference string of a textblock to appear in the column adjacent to the link
    # col 4 is text to appear if the reference cannot be found in the database
    # col 5 normally empty string, if set to text it will replace the textblock

    contents = []

    modules_tuple = editwidget.widget_modules()

    for name in modules_tuple:
        ref = 'widgets.' + name + '.module'
        notfound = 'Textblock reference %s not found' % ref
        contents.append([name, name, '', ref, notfound, ''])

    page_data[("modules","link_table")] = contents


def retrieve_widgets_list(skicall):
    "this call is to retrieve data for listing widgets in a module"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # Fill in header
    call_data['extend_nav_buttons'].append(["list_widget_modules", "Modules", True, ''])

    if 'chosen_module' in call_data:
        module_name = call_data['chosen_module']
    elif 'module' in call_data:
        module_name = call_data['module']
    else:
        raise FailPage("Module not identified")

    modules_tuple = editwidget.widget_modules()

    if module_name not in modules_tuple:
        raise FailPage("Module not identified")

    # set module into call_data
    call_data['module'] = module_name

    page_data[("adminhead","page_head","large_text")] = "Widgets in module %s" % (module_name,)
    page_data[('moduledesc','textblock_ref')] = 'widgets.' + module_name + '.module'

    if 'widgetclass' in call_data:
        # as this page chooses a widget, clear any previous chosen widget class
        del call_data['widgetclass']

    # table of widgets

    # col 0 is the visible text to place in the link,
    # col 1 is the get field of the link
    # col 2 is the get field of the link
    # col 3 is the reference string of a textblock to appear in the column adjacent to the link
    # col 4 is text to appear if the reference cannot be found in the database
    # col 5 normally empty string, if set to text it will replace the textblock

    widget_list = editwidget.widgets_in_module(module_name)
    contents = []
    for widget in widget_list:
        ref = ".".join(("widgets", module_name, widget.classname))
        notfound = 'Textblock reference %s not found' % ref
        classname = widget.classname
        contents.append([classname, classname, '', ref, notfound, ''])

    page_data[("widgets","link_table")] = contents


def retrieve_new_widget(skicall):
    "this call is to retrieve data for displaying a new widget"

    call_data = skicall.call_data
    page_data = skicall.page_data

    # Fill in header
    call_data['extend_nav_buttons'].extend([["list_widget_modules", "Modules", True, ''], ["back_widget_list", "Widgets", True, '']])

    if 'module' not in call_data:
        raise FailPage("Module not identified")

    module_name = call_data['module']
    modules_tuple = editwidget.widget_modules()
    if module_name not in modules_tuple:
        raise FailPage("Module not identified")

    if 'chosen_widget' in call_data:
        widget_class_name = call_data['chosen_widget']
    elif 'widgetclass' in call_data:
        widget_class_name = call_data['widgetclass']
    else:
        raise FailPage("Widget not identified")


    widget_list = editwidget.widgets_in_module(module_name)
    widget_dict = {widgetdescription.classname : widgetdescription for widgetdescription in widget_list}
    
    if widget_class_name not in widget_dict:
        raise FailPage("Widget not identified")

    widg = widget_dict[widget_class_name]
    # widg is a WidgetDescription named tuple

    page_data[("adminhead","page_head","large_text")] = "Create widget of type %s" % (widget_class_name,)


    ref = "widgets." + widg.modulename + "." + widg.classname
    full_textref = ref + '.full'   # the widget full reference string

    adminaccesstextblocks = skilift.get_accesstextblocks(skicall.project)

    if adminaccesstextblocks.textref_exists(full_textref):
        page_data['widgetdesc','textblock_ref'] = full_textref
    else:
        page_data['widgetdesc','textblock_ref'] = ref

    field_contents = []

    for field_argument in widg.fields:
        if field_argument == 'show':
            field_contents.append([field_argument, 'widgets.show'])
        elif field_argument == 'widget_class':
            field_contents.append([field_argument, 'widgets.widget_class'])
        elif field_argument == 'widget_style':
            field_contents.append([field_argument, 'widgets.widget_style'])
        elif field_argument == 'show_error':
            field_contents.append([field_argument, 'widgets.show_error'])
        elif field_argument == 'clear_error':
            field_contents.append([field_argument, 'widgets.clear_error'])
        else:
            field_contents.append([field_argument, ref + '.' + field_argument])

    page_data[('fieldtable','contents')] = field_contents

    if widg.containers:
        page_data[('containerdesc','show')] = True

    # set widget class name into call_data
    call_data['widgetclass'] = widget_class_name

    # display the widget html
    page_data[('widget_code','pre_text')] = widg.illustration


def create_new_widget(skicall):
    "this call is to create and insert a new widget, goes on to widget edit"

    call_data = skicall.call_data
    page_data = skicall.page_data

    project = call_data['editedprojname']

    if 'module' not in call_data:
        raise FailPage("Module not identified")
    if 'widgetclass' not in call_data:
        raise FailPage("Widget not identified")
    if 'location' not in call_data:
        raise FailPage("Location of new widget has not been understood")

    if 'new_widget_name' not in call_data:
        raise FailPage("Widget name missing")
    new_name=call_data['new_widget_name']
    if not new_name:
        raise FailPage("Invalid name")
    new_lower_name = new_name.lower()
    if (new_lower_name == 'body') or (new_lower_name == 'head') or (new_lower_name == 'svg')  or (new_lower_name == 'show_error'):
        raise FailPage(message="Unable to create the widget, the name given is reserved")
    if _AN.search(new_name):
        raise FailPage(message="Invalid name, alphanumeric and underscore only")
    if new_name[0] == '_':
        raise FailPage(message="Invalid name, must not start with an underscore")
    if new_name.isdigit():
        raise FailPage(message="Unable to create the widget, the name must include some letters")

    if 'new_widget_brief' not in call_data:
        raise FailPage("Widget description missing")
    new_brief = call_data['new_widget_brief']

    if 'page_number' in call_data:
        try:
            call_data['pchange'], new_location =  editwidget.create_new_widget_in_page(project,
                                                                         call_data['page_number'],
                                                                         call_data['pchange'],
                                                                         call_data['location'],
                                                                         call_data['module'],
                                                                         call_data['widgetclass'],
                                                                         new_name,
                                                                         new_brief)
        except ServerError as e:
            raise FailPage(e.message)
        call_data['widget_name'] = new_name
        call_data['status'] = "Widget created"
        return

    if 'section_name' in call_data:
        try:
            call_data['schange'], new_location =  editwidget.create_new_widget_in_section(project,
                                                                            call_data['section_name'],
                                                                            call_data['schange'],
                                                                            call_data['location'],
                                                                            call_data['module'],
                                                                            call_data['widgetclass'],
                                                                            new_name,
                                                                            new_brief)
        except ServerError as e:
            raise FailPage(e.message)
        call_data['widget_name'] = new_name
        call_data['status'] = "Widget created"
        return

    raise FailPage("Either a page or section must be specified")


